#! /usr/bin/env python
#
#  GUI
#
#  Created by Jonas Kylling
"""
GUI for uvspec

To run the GUI you need
    python 2.5 or 2.6. Older versions might work
    wxPython 2.8

Stdout and stderr are redirected to .UVSPEC_ERR. This can be turned on and off by setting redirect to True or False.

Plotting
The GUI should be able to plot most output file made with uvspec. Special care
is needed when plotting output files with the umu, zout and phi options. The GUI
uses the number of values in umu, phi and zout to parse the output file. 
If you want to plot an output file from a previous run, you have to open the inputfile
your created the output file with(or type in the same number of values in umu, phi
and zout).

Latest changes:
    Stdout and stderr are redirecte to .UVSPEC_ERR
    Lots of undocumented changes
    Rewrote widgets.py, much more reuse of code
    Changed the layout
    Possible with comments in .OUT files and similar data file. It is probably possible to plot .dat files.
    Moved create_list to create.py, this file might be renamed
    Minor change in OnRun.
    Removed Main.SaveMessage. Use FileDialog.
    Colour support for the info on every widget
    Finally fixed OnOff bug. It was a problem with the SetValue method.
    You now only have to chose .OUT file once and then you can simply hit the run button!
        Now the GUI is much more funnt to use!:-)
    Changed SaveButtons to be SaveRunButtons
    Changed PlotPanel.ReadColumn. No numpy anymore!!:-D
    Added a splash screen
    Hopefully fixed the OnOff bug
    Changed the creation of the design, added a new NoteBook class
    Added plot class and removed the plot list. Plot is no longer a "create_list"
    Added four space identation instead of tabs
"""

#import pdb

import os
import sys
import re
import time
import threading
import subprocess
import pickle


try:
    import wx
    import wx.lib.scrolledpanel
except:
    print "\nError occured while importing wx!"
    print "Quitting..."
    raise
    
if wx.__version__ != "2.8.9.1":
    print "\nYou do not have the same wx version as the developer! The GUI might not work as expected:-)\n"
    print "Your version of wx is %s." % wx.__version__

try:
    import widgets
    import create
except:
    print "\nError occured while importing widgets or create!"
    print "Quitting..."
    raise

# Globals
redirect = False # turn on off redirection of stdout and stderr
redirect_file = ".UVSPEC_ERR" # stdout and stderr are redirected to this file

size = (1000, 450) # window size


def Load_Documentation():
    """
    Load the documentation from GUI_DOC
    The documentation is saved as a dictionary
    """
    try:
        doc_file = open("resources/GUI_DOC", "r")
        Documentation = pickle.load(doc_file)
        doc_file.close()
    except IOError:
        print "Failed to load doc file GUI_DOC.\a"
        Documentation = {}
    # dictionary with needed keys that are normally not in GUI_DOC
    not_set = {"Input File": "Input File",
               "Output File": "Output File",
               "Plot_Out":"Plot a data file"}
    for name, value in not_set.items():
        try:
            Documentation[name]
        except KeyError:
            Documentation[name] = value
        
        Documentation["Output File"] = "Output File"
        Documentation["Plot_Out"] = "Plot a data file"

    return Documentation

Documentation = Load_Documentation()

def SplashScreen():
    """
    Display a splashscreen
    """
    try:
        if not os.path.exists("resources/splash.png"):
            print "Could not find splash.png"
            return
        image = wx.Image("resources/splash.png", wx.BITMAP_TYPE_ANY)
        image = image.ConvertToBitmap()
        splashstyle = wx.SPLASH_CENTRE_ON_SCREEN|wx.SPLASH_NO_TIMEOUT
        style = wx.SIMPLE_BORDER#|wx.STAY_ON_TOP
        SplashWindow = wx.SplashScreen(image, splashstyle, -1, parent=None)
        return SplashWindow
    except Exception, error:
        print "Could not create splash screen.'"
        print error

FileDialog = widgets.FileDialog
ErrorMessage = widgets.ErrorMessage

class Start(wx.Panel):
    """The first panel
    
    Events for the buttons in this class are bound in the main class.
    The WelcomeText might be changed with html drawing for a better look
    """
    def __init__(self, parent):
        """
        Takes only one argumet, parent window
        """
        wx.Panel.__init__(self, parent, style=wx.SIMPLE_BORDER)
        self.WelcomeText = wx.StaticText(self, -1, """Welcome to the uvspec Graphical User Interface.
        
To start a new input file, push the New button.\n\n To open an old input file, push the Open button.

Mandatory options are in red, the rest are optional.        
        
Find more information about uvspec and libRadtran at www.libradtran.org.""", style=wx.ALIGN_CENTRE)

        self.New  = wx.Button(self, -1, "New")
        self.Open = wx.Button(self, -1, "Open")

        self.HorSizer = wx.BoxSizer(wx.VERTICAL)
        self.VerSizer = wx.BoxSizer(wx.HORIZONTAL)
        
        self.VerSizer.Add(self.New, 0, wx.ALL|wx.CENTER, border=20)
        self.VerSizer.Add(self.Open, 0, wx.ALL|wx.CENTER, border=20)
        
        self.HorSizer.Add((1, 1), 1) # Prevent the WelcomeText from crashing into the top of the window
        self.HorSizer.Add(self.WelcomeText, 0, wx.CENTER)
        self.HorSizer.Add(self.VerSizer, 0, wx.CENTER|wx.TOP, border=40)
        self.HorSizer.Add((1, 1), 1)

        self.SetSizer(self.HorSizer)
        
class Page(wx.Panel):
    """The only difference from a normal wx.Panel is that
    this panel comes with its very own Sizer.
    """
    def __init__(self, parent, sizer_flag=wx.VERTICAL, *args):
        """Optional sizer_flag, default is wx.VERTICAL"""
        wx.Panel.__init__(self, parent)
        self.Sizer = wx.BoxSizer(sizer_flag)

class NoteBook(wx.Notebook):
    """Important for layout

    The funcitons CreatePage and CreatePanel is vital for the construciton of the layout
    """
    def __init__(self, parent):
        wx.Notebook.__init__(self, parent, -1, style=wx.NB_TOP)
        
    def AddPages(self, item_list, out):
        """A page should alway be added by calling this function"""
        for page in item_list:
            self.AddPage(*self.CreatePage(page, self, out))

    def CreatePage(self, item_list, parent, out):
        
        max_items = 8 # Max items per column
        
        #main_panel = wx.lib.scrolledpanel.ScrolledPanel(parent, style=wx.VSCROLL)
        main_panel = widgets.StupidScrolledWindow(parent, style=wx.VSCROLL)
        main_panel.Sizer = wx.BoxSizer(wx.HORIZONTAL)

        label = item_list[0][0]
        
        items = item_list[1:] # Everything except the panel label
        item_len = len(items)
        
        if item_len > max_items:
            times, left = divmod(item_len, 2)
            panel, option_dict = self.CreatePanel(main_panel, items[:times + left])
            out.update(option_dict)
            panel2, option_dict = self.CreatePanel(main_panel, items[times + left:])

            main_panel.Sizer.Add(panel, 1, wx.RIGHT, border=10)
            main_panel.Sizer.Add(panel2, 1,)
        else:
            panel, option_dict = self.CreatePanel(main_panel, items)
            
            main_panel.Sizer.Add(panel, 1, wx.EXPAND)
            
        out.update(option_dict)
        
        main_panel.SetSizer(main_panel.Sizer)
        main_panel.SetAutoLayout(1)
        main_panel.SetupScrolling()
        
        return main_panel, label
    
    def CreatePanel(self, parent, items):
            out = {}
            panel = Page(parent)

            largest_string = max([panel.GetTextExtent(string)[0] for string in [item[0]["name"] for item in items]])

            first = items[0][0]["name"]

            for item in items:
                if not redirect and 0: # don't clutter up the log files
                    print "  %s" % item[0]["name"], item[1].__name__  # REMOVE! Watch the progress when the GUI is initalizing
                item[0]["border"]= largest_string - panel.GetTextExtent(item[0]["name"])[0] + 5
                object = item[1](panel, **item[0])
                try:
                    object.SetToolTipString(Documentation[item[0]["name"]])
                except KeyError:
                    print "No documentation for %s." % repr(item[0]["name"])
                object.end_init()
                
                if (type(object) == widgets.OpenButton or type(object) == widgets.SaveRunButtons
                    or type(object) == widgets.PlotPanel or type(object) == widgets.Text or type(object) == widgets.ExpandableOnOff): # These classes have to expand
                    panel.Sizer.Add(object, 0, flag=wx.ALL|wx.EXPAND, border=5)
                elif (type(object) == widgets.Html):
                    panel.Sizer.Add(object, 1, wx.EXPAND|wx.ALL, border=5)
                else:
                    panel.Sizer.Add(object, 0, flag=wx.ALL|wx.EXPAND, border=5)
                # Check if the object has an event handler
                if item[-1]:
                    object.Bind(item[-1][0], item[-1][1])
                out[item[0]["name"]] = object
            out[first].SetFocusFromKbd() # because of the SpinCtrl feature/BUG!!
            panel.SetSizer(panel.Sizer)
            return panel, out

class Main(wx.Frame):
    """Main Frame

    WARNING, the documentation is mostly oudated.
    
    The Main class is a subclass of wx.Frame.
    It is the only top level window of the GUI(some wx.FileDialogs might
    also accidentically become top level window, but this will probably
    not lead to any problems).
    The main frame contains most of the functionality needed for saving,
    opening, running and plotting. Most of the design is also created here.

    Some info about the different dictionaries and Options.
    The self.Options dictionary will after the Main.__init__ method have
    completed, contain all options, such as albedo, wavelength etc.
    The Other dictionary contains everything that don't fit inside
    the Options dictionary, like input_file, out_file and classes
    that probably don't really need to be stored in a dicitonary, like run and Plot.
    The self.Other and self.Options dicitonaries are very important for the GUI
    The tuples plot, general and create_list contains information about how
    the different options and various gui elements will be added. The "format"
    of the tuples with information is a little bit caothic to understand,
    especially for text editors without higlightning of bracket pairs(like mine!).
    ((Page_Name), ({This dictionary will contain different arguments to the
    specific widget}, this_will_be_the_widget_type,
    (wx.OPTIONAL_EVT_TO_BIND_IN_THE_WIDGET, self.EVENT_HANDLER)))
    Every page starts with a tuple which contains the name of the page, and then
    none or more other tuples containing the different other options.
    The options in plot and general will be saved in self.Other
    The options in create_list will be saved in self.Options
    At the moment the page, the parent of every widget/option is not stored anywhere.
    To add a new option simply copy for instance the line concerning albedo.
    Change the needed arguments in the tuple and maybe the widget type.
    More kinds of widgets will be added later in the widgets.py file
    The behaviour of each widget can be seen by running widgets.py
    """
    global Options
    Options = {}    # self.Options contains all options
    global Other
    Other = {} # "old_inp" and "old_order" will contain the old order
                                           # of .INP files that have been read
    Saved = True # Start as saved, changes if anything is changed
    def __init__(self):
        #-----------------------------------------------------------------------
        # Be careful with the brackets when editing the lists
        #plot = ( 
        #    ( ("Plot",),
        #        ({"name":"Plot", "info":"Plot", "ButtonLabel":"Plot"}, widgets.InfoButton, (wx.EVT_BUTTON, self.OnPlot)),
        #        ({"name":"out_file", "format":"*.OUT"}, widgets.OpenButton, ()),
        #        ({"name":"plot_choices"}, PlotChoices, ()),
        #        ),
        #        )
        #-----------------------------------------------------------------------
        general = (
            ( ("General",),
              ({"name":"Input File", "without_clear_button":True}, widgets.SaveRunButtons, (wx.EVT_BUTTON, self.OnSave)),
                # Event binding of the SaveButtons takes place later
              ({"name":"Output File","SaveButtonLabel":"Run", "format":".OUT", "SaveAsButtonLabel":"Save and Run", "without_clear_button":True}, widgets.SaveRunButtons, (wx.EVT_BUTTON, self.OnRun)),
              ({"name":"Plot_Out", "Other1":Other, "Options1":Options, "info":"Output File", "format":"*.*", "choices":["direct irradiance (red)", "diffuse downward irradiance (green)", "upward irradiance (blue)", "radiances"], "without_clear_button":True}, widgets.PlotPanel, (),),
              ({"name":"How To", "filename":"resources/how_to.html"}, widgets.Html, ()),
              ({"name":"new", "info":"New Input File", "ButtonLabel":"New", "without_clear_button":True}, widgets.InfoButton, (wx.EVT_BUTTON, self.OnNew)),
                ),
                )
        #-----------------------------------------------------------------------
        create_list = create.create_list # Edit create.py when addin new options
        #-----------------------------------------------------------------------
        wx.Frame.__init__(self, None, title="GUI for libRadtran", size=size)

        self.error_txt = "" # most errors should be collected here before they are displayed,
                                # it is better with only one long error message than ten small ones

        global tmp_panel
        tmp_panel = wx.Panel(self)
        tmp_panel.Show(False)
        
        self.Start = Start(self)
        self.Edit = NoteBook(self) # Edit contains all the pages

        self.Edit.AddPages(general, Other)    # Create the General tab

        self.Edit.AddPages(create_list, Options) # Create all the other tabs
        
        self.Edit.Show(False)# Only show Edit after either OnNew or OnOpenExistingINP have been run
        
        self.HorSizer = wx.BoxSizer(wx.VERTICAL)
        self.HorSizer.Add(self.Start, 1, wx.EXPAND)
        
        self.SetSizer(self.HorSizer)

        self.Center(True)
        self.SetSizeHints(*size)


        self.Show(True)

        self.Clear() # Because of the intspinner
        
        # Bind some special events that probably only happen once
        self.Bind(wx.EVT_BUTTON, self.OnNewInp, self.Start.New)
        self.Bind(wx.EVT_BUTTON, self.OnOpenExistingINP, self.Start.Open)
        self.Bind(wx.EVT_CLOSE, self.OnExit, self)

    def Has_Change(self, event):
        """Mark that an option has changed and the document needs to be saved"""
        object = event.GetEventObject().GetParent()
        if object:
            object.Has_Changed = True
            self.Saved = False
        event.Skip()

    def OnNew(self, event):
        if self.GuardianAngel():
            self.Clear()
            self.HorSizer.Detach(self.Edit) # Remove the start panel
            self.Edit.Show(False)
        
            self.HorSizer.Insert(0, self.Start, 1, wx.EXPAND)   # Insert the editing panel
            # might be changed with wx.Sizer.Prepend
            self.Start.Show(True) # Make it visible
        
            self.HorSizer.Layout()
            self.Refresh()      # To prevent some ugly redraw errors
        
    def Clear(self):
        #print "#"*40 + " CLEARED! " + "#"*40
        for option in Options:
            Options[option].Clear()
        for special in Other:
            Other[special].Clear()
        widgets.Saved = True
    
    def OnNewInp(self, event):
        """When the New Button is pressed, self.Main.Start.New"""
        tmp = FileDialog(format="*.INP", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if tmp:
            Other["Input File"].SetValue(tmp)
            self.Start_End()
    def OnOpenExistingINP(self, event):
        """When the Open Button is pressed, self.Main.Start.Open"""
        tmp = FileDialog(format="*.INP")
        if tmp:
            Other["Input File"].SetValue(tmp)
            data = self.ReadFile(Other["Input File"].GetValue())
            self.Attribute(data)
            self.Start_End()
            # CHANGE! Set path in edit.info
    def Start_End(self):
        """Removes the start panel and inserts the Edit panel"""
        self.HorSizer.Detach(self.Start) # Remove the start panel
        self.Start.Show(False)                      # and the destroy it
        
        self.HorSizer.Insert(0, self.Edit, 1, wx.ALL|wx.EXPAND, border=5)   # Insert the editing panel
        # might be changed with wx.Sizer.Prepend
        self.Edit.Show(True) # Make it visible
        
        self.HorSizer.Layout()
        self.Refresh()      # To prevent some ugly redraw errors

    def OnExit(self, event):
        """This event is triggered on exit
        
        It might be bound later to at least three events.
        wx.EVT_CLOSE  When the fram is closed
        wx.EVT_BUTTON When the quit button is pressed(quit button will be added later)
        wx.EVT_MENU  When quit is chosen in the Menu
        
        The event should probably have som check if anything have been edited
        since last save and only prompt the user for a save dialog then
        """
        if widgets.Saved:
            #event.Skip()
            self.Destroy()
            return # Do not prompt the user if no change has occured since last save
        else:
            if self.GuardianAngel():
                event.Skip()
    def GuardianAngel(self):
        """ Protect the user. """
        Dialog = wx.MessageDialog(None, "Save Input file before exit?", "Save",
                        wx.YES_NO|wx.CANCEL|wx.YES_DEFAULT)
        answer = Dialog.ShowModal()
        if answer == wx.ID_YES:
            save_file = Other["Input File"].GetValue()
            if not save_file:
                save_file = FileDialog(format=".INP", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
                Other["Input File"].SetValue(save_file)
            self.Save(save_file)
            Dialog.Destroy()
        elif answer == wx.ID_CANCEL:
            Dialog.Destroy()
            return False
        return True

    def OnSave(self, event):
        """Event handler triggered when save button is pressed"""
        tmp = event.GetEventObject().GetParent().GetValue()
        #tmp = FileDialog(".OUT", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if not tmp:
            tmp = FileDialog(".OUT", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
            if not tmp: return
            event.GetEventObject().GetParent().SetValue(tmp)
        self.Save(tmp)
        widgets.Saved = True
    def OnRun(self, event):
        """Event handler triggered when run buttin is pressed
        
        The Run button might be disabled when uvspec is running
        """
        out_file = Other["Output File"].GetValue() # The Output File
        if not out_file:
            out_file= FileDialog(".OUT", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
            if not out_file: return
            Other["Output File"].SetValue(out_file)
        # CHANGE! Thefile should be some kind of safe tempoary file from the standard library tempfile
        tmp_file = os.path.abspath(".tmp_UVSPEC.INP")
        tmp_save = self.Saved
        self.Save(tmp_file) # Save a tempoary INP file with the latest modifications
        self.Saved = tmp_save # Set it to the previous value, since self.Saved sets it to True

        win = widgets.RunUvspec(self, tmp_file, out_file)
        Other["Plot_Out"].SetValue(out_file)
    def env_check(self):
        """Check if the environment variable and uvspec exists
        
        If the environment varaible "LIBRADTRAN_DATA_FILES" don't exist and uvspec 
        can't be found in the path the run button will be disabled.
        """
        print "env_check"
        if not os.getenv("LIBRADTRAN_DATA_FILES"):
            warning = wx.MessageDialog(None, 
            """The Environmentvariabel LIBRADTRAN_DATA_FILES was not found!
            You can still edit .INP files and plot .OUT files, but you can not run UVSPEC! It is strongly recomended to set the environment variabel and restart the GUI!""",
            "Warning!", # Title
            wx.OK)      # Style
            warning.Center(True)
            warning.ShowModal()
            # Disable run Button
            Other["Output File"].Enable(False)
        try:
            subprocess.Popen(("uvspec", "-v"), stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            # remove all output the command
        except OSError:
            warning = wx.MessageDialog(self, 
            """Could not find UVSPEC in your PATH!
            You can still edit .INP files and plot .OUT files, but you can not run UVSPEC! It is strongly recomended to change your PATH and restart the GUI!""",
            "Warning!", # Title
            wx.OK)      # Style
            warning.Center(True)
            warning.ShowModal()
            # Disable run Button
            Other["Output File"].Enable(False)
    def Save(self, path):
        data = ""
        for option in Options:
            if Options[option].Has_Changed:
                values = Options[option].GetValue()
                if values is None: continue
                if isinstance(values, basestring): values = (values,)
                data += ("%s " % option) + " ".join(values) + "\n"
        try:
            print "saving input file to '%s'." % (path)
            print "'%s'" % data
            f = open(path, "w")
            f.write(data)
            f.close()
        except IOError:
            msg = "Could not save input file to '%s'." % (path)
            print msg
            ErrorMessage(msg)
                
    def ReadFile(self, path, includes=[]):
        try:
            f = open(path, "r")
            data = f.readlines()
            f.close()
        except IOError:
            msg = "Could not open %s." % (repr(path))
            self.error_txt += msg
            print " "*len(includes) + msg
            return

        data = [line.split() for line in data]

        # Remove lines with comments and merge continous lines
        line = 0
        while line < len(data):
            if not data[line]: data.pop(line); continue # skip empty lines
            if data[line][0].startswith("#"): data.pop(line); continue # skip comments
            elif [True for word in data[line] if (word.find("#") != -1)]: # remove comments from the line
                tmp = []
                for word in data[line]:
                    pos = word.find("#")
                    if pos != -1:
                        if word != "#":
                            tmp.append(word[:pos])
                        break
                    else:
                        tmp.append(word)
                data[line] = tmp
            elif data[line][-1].endswith("\\"): # continous line
                data[line][-1] = data[pos][-1][:-1] # remove the \
                if data[line][-1] == "": data[line].pop() # if the \ was preceded and continued by whitespace
                data[line].extend(data[line + 1])
            else:
                line += 1

        # Get the includes and include them
        buff = 0
        this_includes = []
        for line in xrange(len(data)):
            if (data[line + buff][0] == "include" and len(data[line + buff]) == 2):
                this_includes.append(data.pop(line + buff)[1])
                buff -= 1

        for include_path in this_includes:
            if not os.path.exists(path):
                msg = "Include %s in %s does not exist.\n" % (include_path, path)
                self.error_txt += msg
                print " " * len(includes) + msg
            elif (this_includes + includes).count(include_path) != 1: # if the file has been included before
                msg = "Include %s included more than one time. One time in %s. Please fix this " % (include_path, path)
                self.error_txt += msg
                print " " * len(includes) + msg
            else:
                include_data = self.ReadFile(include_path, includes + this_includes)
                if include_data: # the includ file might contains errors and return None
                    data.extend(include_data)
        
        if self.error_txt and not includes: # only Display error message in the most outer scope
            ErrorMessage(self.error_txt)
            self.error_txt = ""

        return data

    def Attribute(self, data):
        self.error_txt = ""
        for line in data:
            try:
                error = Options[line[0]].SetValue(line[1:])
                if error: self.error_txt += error + "\n"
            except KeyError:
                error = "Unknown options '%s'.\n" % line[0]
                self.error_txt += error
                print error
        if self.error_txt:
            ErrorMessage(self.error_txt)

def main():
    app = wx.App(redirect=redirect, filename=redirect_file)
    print time.strftime("%Y-%m-%d-%H-%M-%S")
    splash = SplashScreen()
    frame = Main()
    app.SetTopWindow(frame)
    if splash: splash.Close() # close directly after all GUI elements are initialized
    frame.env_check() # have to be called after the Splashscreen is closed
                               # because the MessageDialog and Splashscreen 
                               # will block each other
    """
    #useful when testing the GUI
    Other["Input File"].SetValue("/Users/jonas/tmp.inp")
    data = frame.ReadFile(Other["Input File"].GetValue())
    frame.Attribute(data)
    frame.Start_End()
    Other["Plot_Out"].SetValue("/Users/jonas/tmp.out")
    """
    app.MainLoop()

if __name__ == "__main__":
    main()
