#!/usr/bin/env python
#
#  widgets.py
#  GUI
#
#  Created by Jonas Kylling on 25.02.09.
#  Copyright (c) 2009 __MyCompanyName__. All rights reserved.
#
"""
This file contains most custom widgets used by the UVSPEC GUI

More widgets will be added later.
To see how each widget looks run this file.

Most of the widgets takes many of the same arguments.
Each widget takes a parent Window and a name which will 
be displayed to the left of each widget(except for the 
InfoButton). Each widget contains a SetValue And GetValue
 method. An Enable method might be added later to all 
widgets. At the moment there are no tooltips for the 
widgets, but this might be added later.
"""

import wx
import os
import sys
import time
import itertools
import threading
import subprocess
import webbrowser

#import pdb

from wx.lib import plot
from wx import grid
import wx.html
import wx.lib.customtreectrl
import wx.lib.scrolledpanel
import wx.lib.buttons
import wx.lib.platebtn

import FloatSpin # Made by Andrea Gavana

# Quote from "http://www.wxpython.org/what.php":
# """
# wxPython is a cross-platform toolkit. This means that the same
# program will run on multiple platforms without modification.
# """
# Why do I then need the following code?
if sys.platform == "darwin":
    ClearButton = wx.lib.platebtn.PlateButton
else:
    ClearButton = wx.Button

def FileDialog(format="*.*", flags=wx.FD_OPEN, defaultDir="", defaultFile=""):
    """A simple file dialog
    
    FileDialog(format="") -> path
    
    The function takes an optional format as argument.
    Creates a wx.FileDialog and returns the file that have been chosen.
    """
    if flags&wx.FD_OPEN: title = "Open file"
    else: title = "Save file"
    Dialog = wx.FileDialog(None, title, defaultDir=defaultDir, defaultFile=defaultFile, wildcard=format, style=flags)
    if Dialog.ShowModal() == wx.ID_OK:
        filename=Dialog.GetFilename()
        dirname=Dialog.GetDirectory()
        Dialog.Destroy()
        return os.path.join(dirname, filename)
    Dialog.Destroy()

def OnOrOff(value):
    """ Return on if value is true, else off
    Used by ExpandableOnOff
    """
    if value: return "on"
    else: return "off"

def style_test(style):
    while True:
        yield style

def ErrorMessage(message="", title="Error", style=wx.OK|wx.CENTRE):
    msg = wx.MessageDialog(None, message=message, caption=title, style=style)
    msg.ShowModal()

class Option(wx.Panel):
    def __init__(self, parent, name="", border=0, colour=wx.Colour(0, 0, 0, 255),
                 style=0, without_clear_button=False):
        wx.Panel.__init__(self, parent, style=style)

        self.parent = parent
        self.without_clear_button = without_clear_button
        
        self.Has_Changed = False
        self.name = name

        self.SizerList = []
        self.EnableList = [] # list of objects that will be affected by calls to instance.Enable
        
        self.Info = wx.StaticText(self, -1, name)
        self.Info.SetForegroundColour(colour)

        if not self.without_clear_button:
            self.ClearButton = ClearButton(self, -1, "Reset")
            self.ClearButton.Show(False)

            self.Bind(wx.EVT_BUTTON, self.OnClearButton, self.ClearButton)
        
        
        self.AddToSizer(self.Info, flag=wx.RIGHT|wx.CENTRE, border=border)
    def end_init(self):
        """ Must be called from the outside of the class, when all init methods of any childrens are finished """
        self.Sizer = wx.BoxSizer(wx.HORIZONTAL)

        if not self.without_clear_button:
            self.AddToSizer((0, 0), 1,)
            self.AddToSizer(self.ClearButton, 0, flag=wx.ALIGN_RIGHT|wx.CENTRE)
        
        for obj in self.SizerList:
            self.Sizer.Add(*obj[0], **obj[1])
        self.SetSizer(self.Sizer)

    def AddToSizer(self, *args, **kwargs):
        """
        Method to be called in the same way as wx.Sizer.Add
        Each object should be addetd trough this method and
        not with self.Sizer.Add
        """
        self.SizerList.append((args, kwargs))
        self.EnableList.append(args[0]) # Since most objects will be enabled/disabled
    def Clear(self,):
        self.SetValue(None)
        self.Has_Changed = False
        if not self.without_clear_button:
            self.ClearButton.Show(False)
            self.Layout()
    def OnClearButton(self, event):
        self.Clear()
    def InsertToSizer(self, pos, *args, **kwargs):
        self.SizerList.insert(pos, (args, kwargs))
        self.EnableList.append(args[0])
    def Enable(self, enable=True):
        for obj in self.EnableList:
            obj.Enable(enable)
    def Change(self, event):
        #print self.name, event
        self.Has_Changed = True
        global Saved
        Saved = False

        if not self.without_clear_button:
            self.ClearButton.Show(True)
            self.Layout()
            
        event.Skip()
    def Changed(self,):
        self.Has_Changed = True
        if not self.without_clear_button:
            self.ClearButton.Show(True)
            self.Layout()
    def SetToolTipString(self, tooltip):
        wx.Panel.SetToolTipString(self, tooltip)
        self.Info.SetToolTipString(tooltip)

class Spinner(Option):
    """
    Parent class of all spinner objects, IntSpinner and FloatSpinner
    """
    def SetValue(self, value):
        if value:
            self.Changed()
            count = 0
            if len(value) <= len(self.Spinners):
                # if the number of spinners are unequal the number of values
                for val in value:
                    self.Spinners[count].SetValue(self.type(val))
                    count += 1
        else:
            for spin in self.Spinners:
                spin.SetValue(0)
                
    def GetValue(self):
        out = []
        if self.Has_Changed:
            for spin in self.Spinners:
                out.append(str(spin.GetValue()))
            return out
        
class FloatSpinner(Spinner):
    def __init__(self, parent, name="", spins=[], value=0, max_val=None, min_val=None, increment=1.0, **kwargs):
        Option.__init__(self, parent, name=name, **kwargs)

        self.type = float

        if not spins: # If there are no spins the name will be the only spinner
            spins = [name]

        self.Spinners = []

        for label in spins:         
            spin = FloatSpin.FloatSpin(self, value=value, max_val=max_val,
                                       min_val=min_val, increment=increment)
            
            tmp_sizer = wx.BoxSizer(wx.VERTICAL)
            
            if len(spins) > 1:
                text = wx.StaticText(self, -1, label)
                tmp_sizer.Add(text, flag=wx.CENTER|wx.BOTTOM, border=2)

            tmp_sizer.Add(spin, flag=wx.CENTER|wx.RIGHT, border=10)
            
            self.AddToSizer(tmp_sizer, flag=wx.CENTER)
            
            self.Spinners.append(spin)
        self.Bind(wx.EVT_SPINCTRL, self.Change)

class IntSpinner(Spinner):
    def __init__(self, parent, name="", spins=[], value=0, max_val=1000000000, min_val=-1000000000, **kwargs):
        Option.__init__(self, parent, name=name, **kwargs)

        self.type = int
        
        if not spins: # If there are no spins the name will be the only spinner
            spins = [name]
            
        if max_val is None: max_val = 1000000000
        if min_val is None: min_val = -1000000000
            
        self.Spinners = []

        for label in spins:         
            spin = wx.SpinCtrl(self, value=str(value), max=max_val,
                                       min=min_val)
            spin.SetFocus() # Because wx.SpinCtrl is the ONLY widget which generates
                                   # an event when it loses focus the FIRST time
                                   # It creates ugly jumping on scrolled panels, if the focus is out
                                   # of sight, at least on macs.
            
            tmp_sizer = wx.BoxSizer(wx.VERTICAL)
            
            if len(spins) > 1:
                text = wx.StaticText(self, -1, label)
                tmp_sizer.Add(text, flag=wx.CENTER|wx.BOTTOM, border=2)

            tmp_sizer.Add(spin, flag=wx.CENTER|wx.RIGHT, border=10)
            
            self.AddToSizer(tmp_sizer, flag=wx.CENTER)
            
            self.Spinners.append(spin)
        self.Bind(wx.EVT_SPINCTRL, self.Change)

class FloatSpinTogether(FloatSpinner):
    """The FloatSpins will spin together
    
    Same as FloatSpinner except that an event is added that
    prevent the first SpinCtrl to be higer than the last
    """
    def __init__(self, *args, **kwargs):
        FloatSpinner.__init__(self,*args, **kwargs)
                
        count = 0
        for spin in self.Spinners:
            if count == 2: break
            
            spin.number = count
            self.Bind(wx.EVT_SPINCTRL, self.OnChange, spin)
            
            count += 1
            
    def OnChange(self, event):
        """Prevents the user from setting a higher start than stop value
        
        The value are automatically changed to not be higher or lower than the changing widget.
        """
        object = event.GetEventObject()
        if object.number == 0:
            if self.Spinners[0].GetValue() > self.Spinners[1].GetValue():
                self.Spinners[1].SetValue(self.Spinners[0].GetValue())
        elif object.number == 1:
            if self.Spinners[1].GetValue() < self.Spinners[0].GetValue():
                self.Spinners[0].SetValue(self.Spinners[1].GetValue())
        event.Skip()

class ChoiceObject():
    """
    Contains methods for setting and getting the values of wx.Choice
    Childrens of this class must have a wx.Choice object named self.Choice
    and a list of the different choices named self.choices
    """
    def SetValue(self, value):
        """Set the selected choice"""
        if value:
            if type(value) == list: value = value[0]
            try:
                self.Choice.SetSelection(self.choices.index(value))
                self.Changed()
            except ValueError:
                return "Illegal %s value for %s option." % (repr(value), repr(self.name))
        else:
            self.Choice.SetSelection(0)
    def GetValue(self):
        """Returns the currently selected choice"""
        if self.Has_Changed:
            selected = self.choices[self.Choice.GetCurrentSelection()]
            if selected != "None":
                return self.choices[self.Choice.GetCurrentSelection()]

class Choice(Option, ChoiceObject):
    """A list which the user can chose from
    """
    def __init__(self, parent, choices=[], **kwargs):
        Option.__init__(self, parent, **kwargs)

        self.choices = choices

        self.Choice = wx.Choice(self, -1, choices=self.choices)
        
        self.AddToSizer(self.Choice, flag=wx.CENTER)

        self.Bind(wx.EVT_CHOICE, self.Change)
            
class OnOff(Option):
    """OnOff button"""
    def __init__(self, parent, write_value=True, out_value="", **kwargs):
        Option.__init__(self, parent, **kwargs)
        
        self.out_value = out_value
        self.write_value = bool(write_value)  # Indicate the write state of the button
        
        self.Button = wx.ToggleButton(self, -1, "Off")

        self.AddToSizer(self.Button, flag=wx.CENTER)
        self.SetValue(False)
        
        self.Button.Bind(wx.EVT_TOGGLEBUTTON, self.Change)
        
    def Change(self, event):
        Option.Change(self, event)
        self.ChangeLabel()
        
    def ChangeLabel(self):
        """Fancy changing of the buttonlabel
        Makes it easier to see wether the button is on or off
        """
        if self.Button.GetValue():
            self.Button.SetLabel("On")
        else:
            self.Button.SetLabel("Off")
    def SetValue(self, value):
        """
        Set the value of the button
        Any True value will turn the button on
        """
        if value is None:
            self.Button.SetValue(not self.write_value)
        
        elif value == "on": self.Button.SetValue(True)
        elif value == "off": self.Button.SetValue(False)

        else:
            if value is False: # In case you want to turn the button off
                self.Button.SetValue(not self.write_value)
            else:
                self.Button.SetValue(self.write_value)
                self.Changed()
        self.ChangeLabel()
    def GetValue(self):
        """Returns the value of the button
        
        Because of the behavior of the save function in the GUI
        file the return object is a tuple with an empty string
        """
        if self.Has_Changed:
            value = self.Button.GetValue()
            if value == self.write_value:
                return self.out_value
            else:
                return None

class OpenButton(Option):
    """Let's the user open files and view and edit the path to the currently selected file
    """
    def __init__(self, parent, file="", ButtonLabel="Open", flags=wx.FD_OPEN, format="*.*", **kwargs):
        Option.__init__(self, parent, **kwargs)
        
        self.format = format
        self.flags  = flags

        self.FileInfo = wx.TextCtrl(self, -1 )
        self.FileInfo.SetValue(file)
        
        self.Button = wx.Button(self, -1, ButtonLabel)
        
        self.AddToSizer(self.FileInfo, 20, wx.CENTER|wx.RIGHT, border=10)
        self.AddToSizer(self.Button, 0, flag=wx.CENTER)
        
        self.Button.Bind(wx.EVT_BUTTON, self.Change) # please do not change the order of these event bindings!
        self.Button.Bind(wx.EVT_BUTTON, self.Open) # the order is important for which function get's the event
        
    def SetValue(self, value):
        """Change the value of self.FileInfo"""
        if value:
            if type(value) == str or type(value) == unicode:
                self.FileInfo.ChangeValue(value)
            else:
                self.FileInfo.ChangeValue(value[0])
            self.Changed()
        else:
            self.FileInfo.ChangeValue("")
            self.Changed()
    def GetValue(self):
        """Returns the path to the currently selected file"""
        if self.Has_Changed:
            tmp = self.FileInfo.GetValue()
            if tmp:
                return tmp
    def Open(self, event):
        """FileDialog when the open button is pressed"""
        path = os.path.split(self.FileInfo.GetValue())
        tmp = FileDialog(format=self.format, flags=self.flags, defaultDir=path[0],
                             defaultFile=path[1])
        if tmp:
            self.SetValue((tmp,))
            event.Skip()

class InfoButton(Option):
    def __init__(self, parent, name="", info="info", ButtonLabel="ButtonLabel", **kwargs):
        Option.__init__(self, parent, name=info, **kwargs)
        
        self.name = name
        
        self.Button = wx.Button(self, -1, ButtonLabel)

        self.AddToSizer(self.Button, flag=wx.CENTER)

        self.Button.Bind(wx.EVT_BUTTON, self.Change)

    def SetValue(self, *rubbish, **more_rubbish):
        """No effect and will probably never have any effect"""
        pass
    def GetValue(self):
        """Will probably always return None"""
        return None

class Text(Option):
    """Allows the user to type in whatever he or she wants"""
    def __init__(self, parent, value="", **kwargs):
        Option.__init__(self, parent, **kwargs)
        
        self.Text = wx.TextCtrl(self, -1, value)
        self.AddToSizer(self.Text, 1, flag=wx.CENTER|wx.RIGHT, border=20) # looks ugly without a large border

        self.Bind(wx.EVT_TEXT, self.Change)
        
    def SetValue(self, value):
        """Set the selected choice"""
        if value:
            if not isinstance(value, basestring):
                value = " ".join(value)
            self.Text.ChangeValue(value)
            self.Changed()
        else:
            self.Text.ChangeValue("")
    def GetValue(self):
        """Returns the currently selected choice"""
        if self.Has_Changed:
            return self.Text.GetValue()

class ListAndFile(OpenButton, ChoiceObject):
    def __init__(self, parent, choices=[], *args, **kwargs):
        OpenButton.__init__(self, parent, *args, **kwargs)

        self.choices = choices
        
        self.Choice = wx.Choice(self, -1, choices=self.choices)

        self.InsertToSizer(1, self.Choice, flag=wx.CENTER)
    def GetValue(self):
        return ChoiceObject.GetValue() + OpenButton.GetValue(self,)
    def SetValue(self, value):
        if not value: value = None, None
        if isinstance(value, basestring) or len(value) == 1 :
            OpenButton.SetValue(self, value)
        else:
            ChoiceObject.SetValue(self, value[0])
            OpenButton.SetValue(self, value[1])

class ExpandableOnOff(Option, wx.lib.customtreectrl.CustomTreeCtrl):
    def __init__(self, parent, choices=[], **kwargs):
        Option.__init__(self, parent, **kwargs)
        
        self.tree = wx.lib.customtreectrl.CustomTreeCtrl(self, style=wx.TR_DEFAULT_STYLE)
        #self.tree.SetBackgroundColour("GREY")
        self.boxes = {}
        
        root = self.tree.AddRoot("")
        
        for label in choices:
            box = self.tree.AppendItem(root, label, 1)
            self.boxes[label] = (box)
            self.CheckBox(label, True)
        self.AddToSizer(self.tree, 1, )
        self.Bind(wx.lib.customtreectrl.EVT_TREE_ITEM_CHECKED,
                  self.Change)
        self.Bind(wx.lib.customtreectrl.EVT_TREE_ITEM_CHECKED,
                  self.OnChecked)
    def CheckBox(self, label, value):
        try:
            box = self.boxes[label]
        except KeyError:
            print "%s recived Unknown box %s" % (self.name, label)
            return
        self.tree.CheckItem(box, value)
    def SetValue(self, value):
        if value is None:
            for box in self.boxes:
                self.CheckBox(box, True)
                self.tree.SetPyData(self.boxes[box], True)
            self.Changed()
        else:
            if value[0] == "on":
                self.CheckBox(value[1], True)
            else: # value[0] == "off"
                self.CheckBox(value[1], False)
    def GetValue(self):
        keys = self.boxes.keys()
        keys.sort()
        out = []
        for box in keys:
            #print self.GetPyData(self.boxes[box]), OnOrOff(self.GetPyData(self.boxes[box])), self.GetItemText(self.boxes[box])
            out.append((OnOrOff(self.GetPyData(self.boxes[box])), self.GetItemText(self.boxes[box])))
            
        txt = " ".join(out[0]) + "\n"
        for data in out[1:]:
            txt += self.name + " " +  " ".join(data) + "\n"
            
        return txt[:-1] # remove last newline
    
    def OnChecked(self, event):
        box = event.GetItem()
        if self.tree.GetPyData(box):
            self.tree.SetPyData(box, False)
        else:
            self.tree.SetPyData(box, True)
        event.Skip()

class Expandable(wx.Panel):
    def __init__(self, parent, choices=[]):
        wx.Panel.__init__(self, parent)
        
class ExpandableOnOff(Option):
    def __init__(self, parent, choices=[], default=True, **kwargs):
        """
        choices is a list of the labels of the checkboces
        default is the the default value of the checboxes, True or False, (on or off)
        """
        Option.__init__(self, parent, **kwargs)

        self.parent = parent

        self.default = bool(default) # default state of the checkboxes
        
        self.Expanded = False

        self.ExpandButton = wx.BitmapButton(self, -1, style=wx.NO_BORDER)
        
        self.LabelBitmap = wx.Bitmap("resources/Label.png")
        self.SelectedBitmap= wx.Bitmap("resources/Selected.png")

        self.ExpandButton.SetBitmapLabel(self.LabelBitmap)
        #self.ExpandButton.SetBitmapSelected(selected)
        #self.ExpandButton.SetBitmapHover(selected)
        #self.ExpandButton.SetBitmapFocus(selected)
        #self.ExpandButton.SetBitmapDisabled(selected)
        
        self.VerticalSizer = wx.BoxSizer(wx.VERTICAL)
        
        self.AddToSizer(self.ExpandButton, 0)
        self.AddToSizer(self.VerticalSizer, 1, wx.EXPAND)

        self.Boxes = []
        self.Labels = {}

        for label in choices:
            box = wx.CheckBox(self, -1, label)
            box.SetValue(self.default)
            box.Show(False)
            self.Boxes.append(box)
            self.Labels[label] = box
        self.Bind(wx.EVT_BUTTON, self.OnExpand, self.ExpandButton)
        self.Bind(wx.EVT_CHECKBOX, self.Change)

    def OnExpand(self, event):
        if self.Expanded:
            self.ExpandButton.SetBitmapLabel(self.LabelBitmap)
            for box in self.Boxes:
                self.VerticalSizer.Detach(box)
                box.Show(False)
            self.Expanded = False
        else:
            self.ExpandButton.SetBitmapLabel(self.SelectedBitmap)
            for box in self.Boxes:
                self.VerticalSizer.Add(box, 0, wx.LEFT, border=5)
                box.Show(True)
            self.Expanded = True

        self.parent.GetContainingSizer().Layout()
        
    def GetValue(self):
        print self.name
        txt = ""
        for box in self.Boxes:
            if box.GetValue() == self.default:
                txt += self.name + " " + "on" + " " + box.Label + "\n"
            #txt += self.name + " " + OnOrOff(box.GetValue()) + " " + box.Label + "\n"
        return txt[len(self.name) + 1:-1] # cut off the preceding name and the last newline
    
    def SetValue(self, values):
        if values is None:
            for box in self.Boxes:
                box.SetValue(self.default)
        else:
            try:
                if values[0] == "on":
                    self.Labels[values[1]].SetValue(True)
                else:
                    self.Labels[values[1]].SetValue(False)
                self.Changed()
            except KeyError:
                print "Unknown argument %s to %s." % (repr(values[1]), repr(self.Name))


class TextInfo(wx.Panel):
    def __init__(self, parent, name="", text="", border=0, filename=""):
        wx.Panel.__init__(self, parent)
        self.text = wx.StaticText(self, -1, text, style=wx.ALIGN_CENTRE)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.text, 1, wx.EXPAND)
        
        self.SetSizer(sizer)

        if filename:
            f = open(filename, "r")
            self.text.SetLabel(f.read())
            f.close()
    def end_init(self):
        pass

class HtmlWindow(wx.html.HtmlWindow):
    def OnLinkClicked(self, url):
        webbrowser.open(url.GetHref(), new=2, autoraise=1)

class Html(wx.Panel):
    def __init__(self, parent, name="", text="", border=0, filename=""):
        wx.Panel.__init__(self, parent)

        self.html = HtmlWindow(self)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.html, 1, wx.EXPAND)
        
        self.SetSizer(sizer)
        
        if text:
            self.html.SetPage(text)
        if filename:
            self.html.LoadFile(filename)
    def end_init(self):
        pass
    def SetValue(self, value):
        pass
    def GetValue(self):
        return None
    def Clear(self):
        pass
            
class SaveRunButtons(OpenButton):
    """SaveButtons in general page
    
    The design of the savebuttons.
    """
    def __init__(self, parent, SaveButtonLabel="Save", SaveAsButtonLabel="Save As", **kwargs):
        OpenButton.__init__(self, parent, flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT, ButtonLabel=SaveAsButtonLabel, **kwargs)
        
        self.SaveButton = wx.Button(self, -1, SaveButtonLabel)
        self.InsertToSizer(2, self.SaveButton, flag=wx.CENTER|wx.RIGHT, border=10)
        
def BeautyColour(length):
    """
    Generate the colour spectrum of the rainbow. Used by the plot
    to get beautiful colours:-)
    The colours will be incremented by every next call depending
    on the length.
    """
    if length < 255*3:
        mod = 255*3 / length
    elif length == 255*3:
        mod = 1
    else:
        mod = length / 255*3
    colour = [255, 0, 0] # start at the end of the spectrum
    runs = 0
    while runs < length:
        runs += 1
        if colour[0] == 255 and colour[1] < 255:
            colour[1] += mod
        elif colour[1] == 255 and colour[0] > 0:
            colour[0] -= mod
        elif colour[1] == 255 and colour[2] < 255:
            colour[2] += mod
        elif colour[2] == 255 and colour[1] > 0:
            colour[1] -= mod
        else: # self.colour[2] == 255 and self.colour[0] < 255:
            colour[0] += 255
        for i in xrange(3): # 3 is the length of self.colour
            if colour[i] > 255: colour[i] = 255
            elif colour[i] < 0: colour[i] = 0
        yield wx.Colour(*colour)


class PlotWindow(wx.Frame):
    """Creates a simple plot frame
    
    !! This class will probably be rewritten !!
    """
    def  __init__(self, parent, lines,  name="", MinMax=None, title="", xtitle="", ytitle=""):
        wx.Frame.__init__(self, parent, -1, title, size=wx.Size(500, 500))
        x, y = MinMax # Max and min values for x and y axis
        
        panel = wx.Panel(self,  -1)
        self.Plot = plot.PlotCanvas(panel)
        self.Plot.SetInitialSize(size=(500,(500 - 20)))
        
        self.Center()
        self.Show(True)

        self.Plot.SetEnableGrid(True)
        self.Plot.SetEnableLegend(True)
        self.Plot.SetEnableZoom(True)
        self.Plot.SetFontSizeLegend(10)
        self.Plot.SetUseScientificNotation(True)

        graphics = plot.PlotGraphics(lines, name, xtitle, ytitle)
        
        self.Plot.Draw(graphics, xAxis=x, yAxis=y)
        
        self.Bind(wx.EVT_SIZE, self.OnSize, self)
        
    def OnSize(self, event):
        """Reinitalize when the window is resized.

        A smarter and faster resizing method can be used, that
        only reinitalize the window when resizing is finished.
        """
        windowsize = self.GetSize()
        windowsize = windowsize - (0, 20)   # Keep the plot inside the frame
        self.Plot.SetInitialSize(size=(windowsize))

class PlotChoices(wx.Panel):
    """A panel with checkboxes """
    def __init__(self, parent, name="name", choices=[]):
        """__init__(parent window, name, list of checkbox labels)"""
        wx.Panel.__init__(self, parent)
        
        self.box_dict = {}
        self.box_list = []
        self.Sizer = wx.BoxSizer(wx.VERTICAL)
        num = 0
        for choice in choices:
            box = wx.CheckBox(self, -1, choice)
            self.Sizer.Add(box, flag=wx.ALIGN_LEFT)
            self.box_list.append(box)
            self.box_dict[choice] = box
        self.SetSizer(self.Sizer)
    def __getitem__(self, obj):
        return self.box_dict[obj]
    def __iter__(self):
        """!! Last box will not be iterated!! """
        return iter(self.box_list[:-1])
    
class PlotPanel(Option):
    def __init__(self, parent, Other1={}, Options1={}, info="Info", format="*.*", choices=[], **kwargs):
        Option.__init__(self, parent, style=wx.BORDER_SUNKEN, **kwargs)
        
        # Some dirty globals to be removed
        global Other
        Other = Other1
        global Options
        Options = Options1

        # Grammar is important
        if info[0].capitalize() in ["A", "E" "I", "O", "U", "Y"]: information = "Plot an %s" % info
        else: information = "Plot a %s" % info
        
        self.Button = wx.Button(self, label="Plot")
        self.FileInfo = OpenButton(self, name=info, format=format, without_clear_button=True)
        self.FileInfo.end_init()
        self.Choices = PlotChoices(self, choices=choices)

        self.AddToSizer(self.Button, 0, wx.RIGHT|wx.CENTER|wx.TOP|wx.BOTTOM, border=5)
        self.AddToSizer(self.FileInfo, 1, wx.CENTER|wx.RIGHT|wx.EXPAND, border=5)

        self.AddToSizer(self.Choices)
                
        self.Button.Bind(wx.EVT_BUTTON, self.OnPlot)
        
    def OnPlot(self, event):
        """Setting everything up for plotting"""
        path = self.FileInfo.GetValue()
        
        if not path:
            msg = "You have to chose an output file to plot."
            print msg
            ErrorMessage(msg)
            return
        if not os.path.exists(path):
            msg = "Could not find '%s'." % (path)
            print msg
            ErrorMessage(msg)
            return
        
        try:
            data, min_max = self.ReadData(path)
        except TypeError: # catch error when ReadData returns None
            return
        #self.ManyPlot(data, min_max)
        PlotWindow(self, data, MinMax=min_max, title="Irradiance Plot", xtitle="Wavelength (nm)",
                    ytitle="Irradiance")
        
    def ManyPlot(self, data, min_max):
        """ Plot all the points in a seperate window. Just for debug. """
        count = 1
        for points in data:
            PlotWindow(self, (points,), MinMax=min_max, title="Plot %i" % count, xtitle="Wavelength (nm)",
                       ytitle="Irradiance")
            count += 1
            
    def ReadData(self, path):
        try:
            f = open(path, "r")
            data = f.readlines()
            f.close()
        except IOError, error:
            ErrorMessage("Could not open file %s." % (path))
            print error
            
        try:
            for pos in xrange(len(data)):
                data[pos] = data[pos].split()
                if data[pos][0].startswith("#"): continue
                data[pos] = [float(num) for num in data[pos]]
        except ValueError, error:
            ErrorMessage("Invalid characters in %s. Your file might be corrupted." % (path))
            print error

        if not data:
            ErrorMessage("Empty file.")
        
        if (Options["umu"].GetValue() and Options["phi"].GetValue()):
            return self.ReadBlocks(data)
        elif [True for box in self.Choices if box.IsChecked()]:
            return self.ReadColumns(data)
        else:
            ErrorMessage("You have to check at least one of the checkboxes.")
        
    def ReadBlocks(self, data):
        """ Read an output file in blocks """

        # structure of each block:
        # line with wavelength data
        # line with degrees
        # line with data
        # line with more data ...

        zout = Options["zout"].GetValue()
        if zout:
            legends = zout.split()
            zout = len(legends)
        if not zout:
            zout = 1 # defaults to one
            legends = ["boa"]

        height = len(Options["umu"].GetValue().split()) + 2 # returns the length of the data
        length =  len(Options["phi"].GetValue().split()) # returns the number of datafields we need.
                                                                        # the two first data fields of each line must be skipped
        blocks = [ data[pos:height + pos] for pos in xrange(0, len(data), height) ] # divide the file into separate blocks
        
        graphs = [] # contains all the PolyLines
        x_bounds = [] # contains min, max of the x axis
        y_bounds = []

        #Parsing of the irrdaiances
        for box in self.Choices:
            if box.IsChecked(): # just test if any of the boxes are checked
                tmp_graphs, min_max = self.ReadColumns([block[0] for block in blocks])
                x_bounds.extend(min_max[0])
                y_bounds.extend(min_max[1])
                graphs.extend(tmp_graphs)
                break

        #Parsing of the radiances
        if self.Choices["radiances"].IsChecked():
            points = [[] for i in xrange((height - 2) * length * zout)] # initilaize a list wich will contain the points
            legend_list =  [[] for i in xrange((height - 2) * length * zout)]# legend for each graph, the legends is "boa umu phi"
            zout_val = 0
            try:
                for block in blocks:
                    wave_length = block[0][0]
                    x_bounds.append(wave_length)
                    c = -2 # explained below
                    for line in block[2:]: # skip the two first lines
                        # pos starts at 2 and the first value is at line[2]
                        # the value on the line is stored in out[i + c]
                        # since out starts at zero h has to be -2 and
                        # incremented with the length for every iteration
                        # over each line in each block. 
                        for pos in xrange(2, length + 2): # skip the two first fields
                            points[zout_val + c + pos].append((wave_length, line[pos]))
                            y_bounds.append(line[pos])
                            if not legend_list[zout_val + c + pos]:
                                legend_list[zout_val + c + pos] = legends[zout_val/(length * (height - 2))],  line[0], block[1][pos - 2]
                        c += length
                    zout_val += length * (height -2)
                    if zout_val == zout * length * (height -2): zout_val = 0
            except IndexError, error:
                ErrorMessage("An error occured while reading the file. Please control that your input file and output file cohere.")
                print error
                return
                
            colour = BeautyColour(len(points))
            graphs.extend([ plot.PolyLine(points[i], legend="%5s %5s %5s" % legend_list[i], colour=colour.next(), width=1) for i in xrange(len(points)) ])
        
        if not graphs:
            ErrorMessage("You have to check at least one checkbox.")
            return
        return (graphs, ( (int(min(x_bounds)), int(max(x_bounds)),),
             (int(min(y_bounds)), int(max(y_bounds)),), ),
             )
            
    def ReadColumns(self, data):
        zout = Options["zout"].GetValue()
        if zout:
            legends = zout.split()
            zout = len(legends)
        if not zout:
            zout = 1 # defaults to one
            legends = ["boa"]

        styles = itertools.cycle((wx.SOLID, wx.DOT, wx.LONG_DASH, wx.SHORT_DASH,
                                  wx.HORIZONTAL_HATCH, wx.VERTICAL_HATCH))
            
        x_bounds = []
        y_bounds = []

        column_names = ["direct", "diff down", "upward"]
        
        column_num = 0 # The column of the out file to print
        columns = [] # list of the columns that will be plotted
        points = []

        legend_list = []
        style_list = []

        # initialize lists of various data
        for box in self.Choices:
            column_num += 1
            if box.IsChecked():
                for i in xrange(zout):
                    points.append([]); legend_list.append([]); style_list.append([])
                columns.append(column_num)

        # parse the data, lots of magic numbers
        c = 0
        for i in xrange(1, zout + 1):
            c += 1
            style = styles.next()
            for line in data:
                if c % zout == 0:
                    x_bounds.append(line[0])
                    for pos in xrange(len(columns)):
                        l_pos = pos + len(columns)*(i - 1) 
                        points[l_pos].append((line[0], line[columns[pos]]))
                        y_bounds.append(line[columns[pos]])
                        if not legend_list[l_pos]:
                            legend_list[l_pos] = "%s %s" % (legends[i - 1], column_names[columns[pos] - 1])
                            style_list[l_pos] = style
                c += 1

        return ([ plot.PolyLine(points[pos], legend=legend_list[pos],
                                colour=["RED", "GREEN", "BLUE"][columns[pos] - 1], width=2, style=style_list[pos]) for pos in xrange(len(points)) ],
                     (
                      ( int(min(x_bounds) ), int(max(x_bounds)) ),
                      ( int(min(y_bounds) ), int(max(y_bounds)) ),
                      )
                )
                # Max and min values of x and y axis
                
    def SetValue(self, value):
        self.FileInfo.SetValue(value)
    def GetValue(self):
        return self.FileInfo.GetValue()

class RunUvspecProcess(threading.Thread):
    def __init__(self, window, input_file, output_file):
        threading.Thread.__init__(self)
        self.window = window
        self.input_file = input_file
        self.output_file = output_file
    def run(self):
        wx.CallAfter(self.window.AddToLog,
                     "Running uvspec with input file: %s \noutput file: %s\n" % (self.input_file, self.output_file))

        exit_value = 0

        process_in = open(self.input_file, "r")
        process_out = open(self.output_file, "w")
        process = subprocess.Popen("uvspec", stdin=process_in, stdout=process_out, stderr=subprocess.PIPE)
        wx.CallAfter(self.window.AddToLog, "Uvspec has started.\n")
        while process.poll() == None: # None inidicates that uvspec is not yet finished
            data = process.stderr.read()
            wx.CallAfter(self.window.AddToLog, data)
            if data.upper().find("ERROR") != -1: # catch errors with return code 0
                exit_value = "Error"
            if wx.CallAfter(self.window.__getattribute__, "should_thread_die"):
                del process
                process_in.close()
                process_out.close()
                wx.CallAfter(self.window.UvspecFinished, "cancel")
                wx.CallAfter(self.window.Close())


        process_in.close()
        process_out.close()
        
        if not exit_value or process.poll() != 0: # correcting uvspec returncodes
            exit_value = process.poll()
        wx.CallAfter(self.window.UvspecFinished, exit_value)
        
class RunUvspec(wx.Frame):
    def __init__(self, parent, input_file, output_file):
        size = 400, 200
        wx.Frame.__init__(self, parent, style=wx.DEFAULT_FRAME_STYLE^wx.SYSTEM_MENU, size=size)

        self.parent = parent
        self.input_file = input_file

        self.finished = False
        self.should_thread_die = False

        wx.SetCursor(wx.StockCursor(wx.CURSOR_WAIT))
        #wx.SetCursor(wx.Cursor("resources/hourglass.png", wx.BITMAP_TYPE_ANY)) # Arve does not like this :-)

        self.Log = wx.TextCtrl(self, -1, "", style=wx.TE_MULTILINE|wx.TE_READONLY)
        self.Cancel = wx.Button(self, -1, "Cancel")
        self.InformativeInformation = wx.StaticText(self, -1, "<-- Out of order")
        self.PretendIAmWorking = wx.Gauge(self,)
        self.PretendIAmWorking.Pulse()

        MainSizer = wx.BoxSizer(wx.VERTICAL)
        HorSizer = wx.BoxSizer(wx.HORIZONTAL)

        MainSizer.Add(self.Log, 5, wx.CENTER|wx.EXPAND)
        HorSizer.Add(self.PretendIAmWorking, 2, wx.EXPAND|wx.ALL|wx.CENTER, border=10)
        HorSizer.Add(self.Cancel, 1, wx.RIGHT|wx.CENTER, border=10)
        HorSizer.Add(self.InformativeInformation, 2, wx.RIGHT|wx.CENTER, border=10)
        MainSizer.Add(HorSizer, 1, wx.EXPAND)

        self.SetSizer(MainSizer)
	self.Layout()

        self.Centre()
        self.SetSizeHints(*size)
        self.Show()

        
        self.parent.Enable(False)

        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.Bind(wx.EVT_BUTTON, self.OnCancel)

        self.thread = RunUvspecProcess(self, input_file, output_file)
        self.thread.start()

        
    def OnClose(self, event):
        # veto if uvspec is still running
        if self.finished:
            self.parent.Enable(True)
            event.Skip()
    def OnCancel(self, event):
        self.should_thread_die = True
        print "The cancel button does not work yet!"
    def AddToLog(self, text):
        self.Log.AppendText(text)
        
    def UvspecFinished(self, exit_value):
        self.finished = True
        self.AddToLog("Uvspec finished with exit value %s\n" % exit_value)
        self.PretendIAmWorking.SetValue(100)
        wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW)) # standard cursor
        del self.thread
        os.remove(self.input_file)

        if exit_value == 0: # success
            self.Close()

class StupidScrolledWindow(wx.lib.scrolledpanel.ScrolledPanel):
    """
    Prevent the scroll windows rom jumping up and down, when a child past
    the window border is focused. 
    """
    def __init__(self, *args, **kwargs):
        wx.lib.scrolledpanel.ScrolledPanel.__init__(self, *args, **kwargs)

    def OnChildFocus(self, *args, **kwargs):
        """ Overwritten to prevent the jumping when a child is focused. """
        pass


def test():
    # Satisfying and readable testcode
    app = wx.PySimpleApp()
    x_size, y_size = 690, 300
    frame = wx.Frame(None, -1, title="Widgets test", size=(x_size, y_size))
    panel = wx.Panel(frame)
    GridSizer = wx.GridSizer(rows=5)    
    for obj in [FloatSpinTogether,  # Add every new item to this list
                FloatSpinner,
                #CheckText,
                OpenButton,
                OnOff,
                #Spinners,
                Choice,
                InfoButton,
                Text,
                SaveRunButtons
                ]:
        if obj is Choice:
            object = obj(panel, name=obj.__name__, choices=["One", "Two", "Three"])
        #elif obj is InfoButton:
        #    object = obj(panel, name=obj.__name__, info=obj.__name__)
        elif obj is FloatSpinner:
            object = obj(panel, name=obj.__name__, spins=["loooong", "3"], increment=0.01)
        elif obj is FloatSpinTogether:
            object = obj(panel, name=obj.__name__, spins=["loooong", "3"], increment=0.01)
        else:
            object = obj(panel, name=obj.__name__,)
        object.end_init()
        GridSizer.Add(object, 1, flag=wx.ALL|wx.CENTER, border=5)
    panel.SetSizer(GridSizer)
    frame.Centre()
    frame.SetSizeHints(x_size, y_size)# If you size it smaller things will start to look ugly
    frame.SetTransparent(220)  # Eyecandy!
    frame.Show(True)
    app.MainLoop()

if __name__ == "__main__":
    test()






"""
Not in use anymore, but worth keeping
"""

class VarFloat(FloatSpinner):
    def __init__(self, parent, max_val=None, min_val=None, increment=0, **kwargs):
        FloatSpinner.__init__(self, parent, max_val=max_val, min_val=min_val,
                              increment=increment, **kwargs)

        self.args = {
            "max_val":max_val,
            "min_val":min_val,
            "increment":increment,
            }
        
        self.NumberOfFloats = 1
        self.FloatCount = wx.SpinCtrl(self, -1, min=1, max=2000000, value="1")
        self.InsertToSizer(1, self.FloatCount, flag=wx.CENTER|wx.RIGHT, border=5)

        self.FloatCount.Bind(wx.EVT_SPINCTRL, self.OnFloatCount)

    def OnFloatCount(self, event):
        num = self.FloatCount.GetValue()
        if num > self.NumberOfFloats:
            for i in xrange(num - self.NumberOfFloats): self.AddFloat()
        elif num < self.NumberOfFloats:
            for i in xrange(self.NumberOfFloats - num): self.RemoveFloat()
        self.NubmerOfFloats = num
        self.Layout()
        self.GetParent().Layout()
        
    def AddFloat(self):
        spin = FloatSpin.FloatSpin(self, **self.args)
        self.Spinners.append(spin)
        self.Sizer.Add(spin, flag=wx.RIGHT|wx.CENTER, border=5)
        self.NumberOfFloats += 1
        
    def RemoveFloat(self):
        obj = self.Spinners.pop()
        self.Sizer.Remove(self.NumberOfFloats + 1)
        self.NumberOfFloats -= 1
        del obj
        
    def GetValue(self):
        return FloatSpinner.GetValue(self)
    
    def SetValue(self, value):
        self.FloatCount.SetValue(len(value))
        self.OnFloatCount(None)
        FloatSpinner.SetValue(self, value)
            
class FloatAndList(FloatSpinner, ChoiceObject):
    def __init__(self, parent, choices=[], **kwargs):
        FloatSpinner.__init__(self, parent, **kwargs)
        self.choices = choices
        self.NumberOfFloats = 1
        self.FloatCount = wx.SpinCtrl(self, -1, min=1, max=2000000, value="1")
        self.Choice = wx.Choice(self, -1, choices=self.choices)
        
        self.InsertToSizer(1, self.FloatCount, flag=wx.CENTER|wx.RIGHT, border=0)
        self.InsertToSizer(1, self.Choice, flag=wx.CENTER)

        self.FloatCount.Bind(wx.EVT_SPINCTRL, self.OnFloatCount)
        self.Bind(wx.EVT_CHOICE, self.Change)

    def OnFloatCount(self, event):
        num = self.FloatCount.GetValue()
        if num > self.NumberOfFloats:
            for i in xrange(num - self.NumberOfFloats): self.AddFloat()
        elif num < self.NumberOfFloats:
            for i in xrange(self.NumberOfFloats - num): self.RemoveFloat()
        self.NubmerOfFloats = num
        self.Layout()
        self.GetParent().Layout()
        
    def AddFloat(self):
        spin = FloatSpin.FloatSpin(self)
        self.Spinners.append(spin)
        self.Sizer.Add(spin, flag=wx.RIGHT|wx.CENTER, border=0)
        self.NumberOfFloats += 1
    def RemoveFloat(self):
        obj = self.Spinners.pop()
        self.Sizer.Remove(self.NumberOfFloats + 2)
        self.NumberOfFloats -= 1
        del obj
        
    def GetValue(self):
        out = FloatSpinner.GetValue(self)
        choice = ChoiceObject.GetValue(self)
        if choice: out.extend((choice,))
        return out
    
    def SetValue(self, value):
        try:
            float(value[-1]) # Trigger a ValueError if the last value is an entry in the choiceobject
            self.FloatCount.SetValue(len(value[:-1]))
            FloatSpinner.SetValue(self, value)
        except ValueError:
            self.FloatCount.SetValue(len(value[:-1]))
            self.OnFloatCount(None)
            FloatSpinner.SetValue(self, value[:-1])
            ChoiceObject.SetValue(self, value[-1])

class VarqFloat(Option):
    def __init__(self, parent, max_val=None, min_val=None, increment=1.0, **kwargs):
        Option.__init__(self, parent, **kwargs)

        ButtonUp = wx.Button(self, -1, "+")
        ButtonDown = wx.Button(self, -1, "-")
        
        self.table = grid.Grid(self, size=(195, 80))
        self.table_init()

        Sizer = wx.BoxSizer(wx.VERTICAL)
        HorSizer= wx.BoxSizer(wx.HORIZONTAL)
        HorSizer.Add(ButtonUp, flag=wx.RIGHT, border=5)
        HorSizer.Add(ButtonDown, flag=wx.ALIGN_RIGHT)
        Sizer.Add(self.table)
        Sizer.Add(HorSizer)

        self.AddToSizer(Sizer)

        ButtonUp.Bind(wx.EVT_BUTTON, self.OnAddFloat)
        ButtonDown.Bind(wx.EVT_BUTTON, self.OnRemoveFloat)

    def OnAddFloat(self, event):
        self.Add_Float(1)
    def OnRemoveFloat(self, event):
        self.Remove_Float(1)
    def Add_Float(self, num):
        start = self.table.GetNumberRows()
        self.table.AppendRows(num)
        end = self.table.GetNumberRows()
        for i in xrange(start, end):
            self.table.SetCellValue(i, 0, "0.0")
    def Remove_Float(self, num):
        if self.table.GetNumberRows() - num > 0:
            self.table.DeleteRows(self.table.GetNumberRows() - 1, num)
        else:
            self.table.DeleteRows(self.table.GetNumberRows()-1, self.table.GetNumberRows()-1) # delete everything except the first value
    def table_init(self):
        self.table.SetDefaultRenderer(grid.GridCellFloatRenderer())
        self.table.SetDefaultColSize(140)
        self.table.SetRowLabelSize(30)
        self.table.SetColLabelSize(20)
        self.table.CreateGrid(1, 1)
        self.table.SetCellValue(0,0, "0.0")
        self.table.SetColLabelValue(0, "Umu")
        
    def OnFloatCount(self, event):
        num = self.FloatCount.GetValue()
        if num > self.NumberOfFloats:
            for i in xrange(num - self.NumberOfFloats): self.AddFloat()
        elif num < self.NumberOfFloats:
            for i in xrange(self.NumberOfFloats - num): self.RemoveFloat()
        self.NubmerOfFloats = num
        self.Layout()
        self.GetParent().Layout()
        
    def GetValue(self):
        out = []
        for i in xrange(self.table.GetNumberRows()):
            out.append(self.table.GetCellValue(i, 0))
        return out
    
    def SetValue(self, value):
        for i in xrange(len(value)): pass
        FloatSpinner.SetValue(self, value)
